% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/earthtide.R
\docType{class}
\name{Earthtide}
\alias{Earthtide}
\alias{Earthtide-class}
\title{Earthtide class}
\format{An \code{\link{R6Class}} generator object}
\description{
Class to generate synthetic earthtide signals.
}
\section{Usage}{

\preformatted{
et <- Earthtide$new(
  utc = as.POSIXct("2017-01-01", tz = "UTC") + 0:(24 * 7) * 3600, 
  latitude = 52.3868,
  longitude = 9.7144,
  catalog = "ksm04",
  wave_groups = data.frame(start = 0.0, end = 6.0))

et$predict(method = "gravity", astro_update = 1)
et$analyze(method = "gravity", astro_update = 1)
et$lod_tide()
et$pole_tide()
et$tide()
et$print()
}
}

\section{Arguments}{

\itemize{
  \item{et: }{An \code{Earthtide} object.}
  \item{utc: }{The date-time in UTC (POSIXct vector).}
  \item{latitude: }{The station latitude (numeric) defaults to 0.}
  \item{longitude: }{The station longitude (numeric) defaults to 0.}
  \item{elevation: }{The station elevation (m) (numeric) defaults to 0.}
  \item{azimuth: }{Earth azimuth (numeric) defaults to 0.}
  \item{gravity: }{Gravity at the station (m/s^2) (numeric) 0 to 
    estimate gravity from elevation and latitude.}
  \item{earth_radius: }{Radius of earth (m) (numeric) defaults to 6378136.3 }
  \item{earth_eccen: }{Eccentricity of earth (numeric) 
    defaults to 6.69439795140e-3}
  \item{cutoff: }{Cutoff amplitude for constituents (numeric) 
    defaults to 1e-6}
  \item{wave_groups: }{Two column data.frame having start and end of 
    frequency groups (data.frame).}
  \item{catalog: }{Use the "hw95s" catalog or "ksm04" catalog(character).}
  \item{...: }{Currently not used.}
  \item{method: }{For \code{predict} and \code{analyze}. One of "gravity", 
    "tidal_potential", "tidal_tilt", "vertical_displacement", 
    "vertical_strain", "areal_strain", "volume_strain", or "ocean_tides".}
  \item{astro_update: }{For \code{predict} and \code{analyze}. Integer that
    determines how often to phases are updated in number of samples. Defaults
    to 1 (every sample), but speed gains are realized with larger values.
    Typically updating every hour will have speed gains and keep precision 
    (ie 3600 for one second data, 60 for minute data, 1 for hourly data).}
}
}

\section{Details}{


\code{$new(utc, latitude, longitude, elevation, azimuth, gravity, 
  earth_radius, earth_eccen, cutoff, wave_groups, catalog, ...)} 
  create a new \code{Earthtide} object and initialize catalog, 
  station and times.

\code{$predict(method, astro_argument)} generate a combined 
  synthetic Earth tide.
  
\code{$analyze(method, astro_argument)} generate components 
  of the Earth tide for analysis.
  
\code{$lod_tide()} generate components of the LOD (Length Of Day) tide.

\code{$pole_tide()} generate components of the pole tide.

\code{$tide()} get the tide \code{data.frame}.

\code{$print()} print the \code{Earthtide} object.
}

\examples{

et <- Earthtide$new(
  utc = as.POSIXct("2017-01-01", tz = "UTC") + 0:(24 * 7) * 3600, 
  latitude = 52.3868,
  longitude = 9.7144,
  catalog = "ksm04",
  wave_groups = data.frame(start = 0.0, end = 6.0))

et$predict(method = "gravity", astro_update = 1)

plot(gravity~datetime, et$tide(), type='l')

}
\references{
Hartmann, T., Wenzel, H.-G., 1995. The HW95 tidal potential catalogue. Geophys. Res. Lett. 22, 3553–3556. \doi{10.1029/95GL03324}

Kudryavtsev, S.M., 2004. Improved harmonic development of the Earth tide-generating potential. J. Geod. 77, 829–838. \doi{10.1007/s00190-003-0361-2}

Wenzel, H.G., 1996. The nanogal software: Earth tide data processing package ETERNA 3.30. Bull. Inf. Marées Terrestres, 124, pp.9425-9439. \url{http://www.eas.slu.edu/GGP/ETERNA34/MANUAL/ETERNA33.HTM}
}
